module Bitmap where

import Int
import Bits hiding (rotate)
import IOExts (unsafePerformIO, trace)
import Addr

import qualified DExp as D
import Exp
import PointE
import Color
import Image
import Arrays
import LoadBitmap

{- BEGIN_GHC_ONLY
import qualified Addr (addrToInt)
   END_GHC_ONLY -}

-- Whether a point is within array bounds.
inBounds :: IntE -> IntE -> Region
inBounds w h = \ (x,y) -> x `under` w &&* y `under` h

under :: FloatE -> IntE -> BoolE
z `under` m = 0 <=* zI &&* zI <=* m - 1 where zI = floorE z

-- ToDo: Have the code generator produce a good representation for nested
-- arrays in which the inner array's size does not depend on the outer
-- array index.  In that case, use contiguous memory.

-- Generalize when we've merged color array construction with others.
-- pixelize :: ImageC -> Array2 ColorE
pixelize :: (IntE,IntE) -> Image a -> Array2 a
pixelize (w,h) img = Array2 w h (t_img . i2fPair)
 where
   t_img = translate (i2fPair (w `div` 2, h `div` 2)) img


-- Bilerp over whole color grid (array).  Do simple bilerp inside each
-- square.  Using f2i instead of floor here, because we can assume x and y
-- are positive. [floor rounds towards minus infinity, f2i used to round toward
-- zero; it now rounds towards minus infinity.]
--
bilerpCArray2 :: ((IntE, IntE) -> ColorE) -> ImageC
bilerpCArray2 sub = \ (x,y) ->
 let
     i = f2i x;  dx = x - i2f i
     j = f2i y;  dy = y - i2f j
 in
     bilerpC (sub (i, j  )) (sub (i+1, j  ))
             (sub (i, j+1)) (sub (i+1, j+1))
             (dx, dy)


-- Bilerp-based bitmap reconstruction.  Account for finiteness.
reconstruct  :: Array2 ColorE -> ImageC
reconstruct (Array2 w h sub) =
  -- Move bitmap's center to the origin.  Use integer div so we'll get an
  -- integral motion even for odd width or height.  Helps speed.
  translate (- i2f (w `div` 2), - i2f (h `div` 2)) cropped
 where
   cropped = crop (inBounds (w-1) (h-1)) bilerped
   -- 
   -- This experimental version splits the one conditional into two, so
   -- that the outer one can float up to the outer loop.  Improves
   -- unrotated spock's pixel rate by about 20%.  However, this is
   -- basically not really a good approach, because rotation and some
   -- other transforms make bitmap x,y depend on both x&y of imported
   -- bitmap.
   cropped' (x,y) = ifE (y `under` (h-1))
                      (ifE (x `under` (w-1))
                        (bilerped (x,y))
                        invisible)
                      invisible
   bilerped = bilerpCArray2 sub

-- Why w-1 and h-1?  Because bilerpCArray2 samples the array at (i2f x + 1)
-- and (i2f y + 1), which must be within (w,h).  However, I think we're
-- sometimes shaving off too much.

importBMP :: String -> Array2 ColorE
importBMP fileName = Array2 width height (sampleBits width)
 where
   (w,h,depth, clrUsed) =
      maybe (error $ "importBMP: Couldn't import " ++ fileName)
            id
	    (getHeader fileName)

   width  = fromInt (toInt w)
   height = fromInt (toInt h)
	    
   sampleBits =
        case depth of
	    32 -> sampleBGR32 bitsAddr  -- high byte is undefined
            24 -> sampleBGR24  bitsAddr
	    8  -> sample8 (bitsAddr + colMapSize 256) (bitsAddr - 1)
	    4  -> sample4 (bitsAddr + colMapSize 16)  (bitsAddr - 1)
            _  -> error $ "importBMP: Depth not supported " ++ show depth

   colMapSize n
	| clrUsed == 0 = n * 4
	| otherwise    = fromInt (toInt clrUsed) * 4

   bitsAddr   = E (D.ImportData fileName)

{- If importing the bmp data at compile-time, use these:
   bitsAddr   = fromInt (addrToInt bitsAddr')
   colMapAddr = fromInt (addrToInt colMapAddr')
   (bitsAddr', colMapAddr', res) = unsafePerformIO (loadBMPBitsPal fileName)
-}

sample8 bA cA w (i,j) = fromBGR24 (deref32 (cA + 4 * cmap_idx))
  where
   cmap_idx = deref8 (bA + w * j + i)

sample4 bA cA w (i,j) = fromBGR24 (deref32 (cA + 4 * cmap_idx))
  where
    -- (`div` 2) the pixel width to get the width in bytes of each
    -- scanline. Do the same for X coord to localise the byte containing
    -- the nibble we're after.
    -- 
   wbytes  = ((w >>> 1) + 3) .&. (-4)
   x       = i >>> 1
   
   off = wbytes * j + x

   cmap_byte = deref8 (bA + off)

    -- The byte fetched contain the nibble in the lower or upper 4 bits.
    -- Select it by looking at whether the LSB is set in the X coord. 
    -- 
   cmap_idx  = (cmap_byte >>> (((i+1) .&. 1) <<< 2)) .&. 0x0f

-- The 32-bit case is easy: Grab a whole 32-bit word at once.
sampleBGR32 bptr w (i,j) =
  -- fromBGR24 (deref32 (bptr + 4 * (w * j + i)))
  -- Separate j from i since they arrive in different phases.  Better
  -- if the compiler did the separation.
  fromBGR32 (deref32 (bptr + 4 * (w * j) + 4 * i))

-- The 32-bit case is easy: Grab a whole 32-bit word at once.
sampleBGRA32 bptr w (i,j) =
  -- fromBGR24 (deref32 (bptr + 4 * (w * j + i)))
  -- Separate j from i since they arrive in different phases.  Better
  -- if the compiler did the separation.
  fromBGRA32 (deref32 (bptr + 4 * (w * j) + 4 * i))

{-
-- The 24-bit case is harder: Grab a byte at a time
sampleBGR24 bptr w (i,j) =
   -- It's an abuse to use bgra here for the values that are
   -- scaled by 255.
   vecRecip255 * bgra (fetch 0) (fetch 1) (fetch 2) 255.0
 where
   -- Width in bytes.  Align to next multiple of four
   wbytes  = (3 * w + 3) .&. (-4)
   addr    = bptr + wbytes * j + 3 * i
   -- Fetch nth byte after addr and left shift (little-endian)
   -- fetch i = deref8 (addr + i) <<< (8 * i)
   -- Instead, fetch and convert to float.
   fetch i = i2f (deref8 (addr + i))
-}

sampleBGR24 bptr w (i,j) =
   -- Grab a (32-bit) double-word, including the blue byte of the next
   -- pixel.  Using fromBGR32 instead of fromBGR32 acknowledges that the
   -- upper byte is undefined.
   fromBGR32 (deref32 addr)
 where
   -- Width in bytes.  Align to next multiple of four
   wbytes = (3 * w + 3) .&. (-4)
   addr   = bptr + wbytes * j + 3 * i

-- Mask to zero out the upper byte in a 32-bit value.
-- bgr_mask = (1 <<< 24) - 1 :: BGR24E


{-
sampleBGR24 bptr w (i,j) = fromBGR24 nCol
 where
   -- Width in bytes.  Align to next multiple of four
   wbytes = (3 * w + 3) .&. (-4)
   nCol   = deref32 (bptr + wbytes * j + 3 * i)
-}


-- Load a BMP header, returning width, height, bit-depth, and # colors used
getHeader :: String -> Maybe (Int32, Int32, Int32, Int32)
getHeader fname =
  -- We don't yet handle compressed or indexed.
  if  flg /= 0 && (planes == 0 || compress == 0 || clrUsed == 0)  then
     -- trace (show (w,h,flg)) $
     Just (w, h, bitCount, clrUsed)
  else Nothing
 where      
  (w,h,flg) = unsafePerformIO (getPicDimensions fname)
  compress = 0 
  clrUsed  = 0
  bitCount = 32
  planes   = 0

{- Old BMP-reader code:
  (w, h, planes, bitCount, compress, clrUsed, clrImp, _)
    = unsafePerformIO (loadBMPHeader fname)
-}
    
importImageSize :: String -> ((FloatE,FloatE), ImageC)
importImageSize fileName = (i2fPair (arr2Size a), reconstruct a)
 where 
   a = importBMP fileName

importImage :: String -> ImageC
importImage = snd . importImageSize

importImageTiled :: String -> ImageC
importImageTiled = uncurry tile . importImageSize

{-  

-- Out for now.  Doesn't work with Color3 (since ColorE is not an Exp),
-- and not the right thing to do in that case.  Should probably use an
-- ARGB32 in any case.

-- Like (chromaKey eps bg (importImage fileName)), but does the chroma
-- matching once at the array level, rather than at each display.  Depends
-- on better hoisting, however, so the new color array is constructed once
-- and assigned to a global, rather than once per frame.
importImageChromaKeyed :: FloatE -> ColorE -> String -> ImageC
importImageChromaKeyed eps bg fileName =
  reconstruct $ chromaArray2 eps bg $ importBMP fileName

chromaArray2 :: FloatE -> ColorE -> Array2 ColorE -> Array2 ColorE
chromaArray2 eps bg arr =
  -- The reifyArray2 line makes a big difference.
  reifyArray2 $  
  arr2Map (\ c -> ifE (withinC eps c bg) invisible c) arr

-}

{- *** Hugs98 has now got Addr.addrToInt (Nov99)  ***
-- Hack: Mag doesn't have Addr, so we use Int.  There doesn't seem to be a
-- way to convert with Hugs, so I rolled my own, by showing the Addr, and
-- parsing the resulting hexadecimal numeral.  Oddly, the same
-- implementation doesn't work in GHC, but fortunately GHC has
-- Addr.addrToInt.
addrToInt :: Addr -> Int
addrToInt addr = 
{- BEGIN_HUGS_ONLY -}
   fst (head (readHex (show addr)))
{- END_HUGS_ONLY -}
{- BEGIN_GHC_ONLY
   Addr.addrToInt addr
   END_GHC_ONLY -}
-}
