module CompileUtils where

import DExp
import Types
import List
import IOExts (trace)

-- How to allocate locals during code generation?  Pass in a "pool" of
-- free (unused) locals.  Each one is an ordinal and a type.  At each
-- "let", search the free local pool for one of the required type.  If
-- none, allocate new local.  In any case, add binding of name to local
-- number in compilation context for the body of the let.  Pass in the
-- number of locals allocated total, for the next allocation.  Let
-- consumes a local, so it is not passed down in pool.  Compile returns
-- the locals now available.

-- Compile an expression in a typing environment Hugs complains if I
-- uncomment the typesig, but it infers the same type....what's going on?

-- Temp:
-- cse = id

-- Typed locals, identified by position in locals list, and by type.
type Local = (Int, Type)

-- Binding of name to local
type BoundLocal = (Id, Local)

-- Procedure argument, identified by position in list.
type Param = (Id, Type)

-- binding of an Exp name to a type and a target language id.
type BoundVar = (Id, (Id, Type))

showTyping :: Id -> Type -> String
showTyping name ty = name ++ " :: " ++ show ty

showBoundLocal :: BoundLocal -> String
showBoundLocal (name, (_, ty)) = showTyping name ty

showParam :: Param -> String
showParam = uncurry showTyping

showSep :: (a -> String) -> [a] -> String
showSep showFun as = concat (intersperse ", " (map showFun as))

-- The free locals and number of locals generated
type AvailLocals = ([Local], Int)

-- Find or make a local of the given type.  Return one and the
-- now-available locals (not containing the specified one).
findOrMakeLocal :: Type
		-> AvailLocals
		-> (Local,AvailLocals)
findOrMakeLocal ty ls = allocLocal (+1) ty ls

-- don't try to reuse, allocate always.
makeLocal :: Type -> AvailLocals -> (Local, AvailLocals)
makeLocal ty (locals, next) = ((next,ty), (locals, next+1))

{-
 Generalise 'findOrMakeLocal' by having it take a function that computes
 the next Int value we associate with a new local slot, in case none can be
 reused. This generalisation is useful when generating code that refers to 
 locals by their offset from a base pointer (on the C stack).
-}
allocLocal :: (Int -> Int)
           -> Type
	   -> AvailLocals
	   -> (Local,AvailLocals)
allocLocal nLocal ty (locals, next) = loop locals []
 where
   -- Search for a matching free local.
   -- No more locals.  Allocate another.
   loop [] _ = -- trace ("made local " ++ show (next,ty) ++ "\n") $
               ((next,ty), (locals, nLocal next))

   -- When we do find one, remove it.  We reverse the predecessor list,
   -- but the order is not important.
   loop (l@(n, lty) : locs') skipped
     | ty == lty  = (l, (skipped ++ locs', next))  -- Found one.  Remove
     | otherwise  = loop locs' (l : skipped)  -- Keep looking

-- Should probably be wrapped up in a monad.
type Compiler m = AvailLocals -> (m (), AvailLocals)


