
{-
  To assist in the prototyping/configuration of various image effects,
  Pan provides a simple UI layer that lets you create UI elements which
  acts as data sources that can be 'plugged' into your image effect.

-}
module UI 
	(
	  UI

	, Label
	, FileName -- aliases for String

	, property          -- :: String -> (Double,Double) -> Double -> UI FloatE
	, slider            -- :: Label -> (Double,Double) -> Double -> UI FloatE
        , islider           -- :: Label -> (Int, Int) -> Int -> UI IntE

	, radioBoxInt       -- :: Label -> [Label] -> Int -> UI IntE
        , radioBox          -- :: Syntactic a => String -> [(Label,a)] -> UI a
	, tickInt           -- :: Label -> Int  -> UI IntE
	, tick              -- :: Label -> Bool -> UI BoolE
	, button            -- :: Label -> Int -> UI ()

	, mkDisplay         -- :: (Displayable a) => (Int,Int) -> a -> UI ()
	, pluginDisplay     -- :: (Displayable a) => (Int,Int) -> a -> UI ()
	, dxtDisplay        -- :: (Displayable a) => (Int,Int) -> a -> UI ()

	, browseImage       -- :: Label -> FileName -> UI ImageC
	, browseImageSize   -- :: Label -> FileName -> UI (FloatE,FloatE,ImageC)
        , imageSourceD      -- :: Label -> UI (Array2 ColorE)
	, imageSource       -- :: Label -> UI ImageC
	, imageSourceSize   -- :: Label -> UI (FloatE, FloatE, ImageC)

	, browseImageWithAlpha -- :: Label -> FileName -> UI ImageC
	, browseImageWithAlphaSize -- :: Label -> FileName -> UI (FloatE,FloatE,ImageC)

        , textBoxD          -- :: String -> UI (Array2 ColorE)
        , textBoxSize       -- :: String -> UI (FloatE, FloatE, ImageC)
        , textBox           -- :: String -> UI ImageC
	, mouseInput        -- :: UI PointE

	, hbox              -- :: UI a -> UI a
	, vbox              -- :: UI a -> UI a
	, topBox            -- :: UI a -> UI a
	, okCancel          -- :: UI a -> UI a

	  -- exports needed by Compiler only:
	, UIState(..)
	, UIElem(..)
	, Backend(..)
	, deUI
	, topBoxLayout
	, mapUIElem         -- :: (UIElem -> UIElem) -> UIElem -> UIElem
	, mapAccumL_UIElem
	
	, fixedH
	, fixedV
	
	, toCId
	, Name

	) where

import Exp hiding (occFreeIn)
import DExp (occFreeIn,substVar)
import qualified DExp (Exp(..))
import Color (ColorE, white, toBGRA32, bgra32ToBGR32 )
import PointE (PointE)
import Image
import Types
import Bitmap
import Arrays
import CCompile ( NameSupply, CCode )
import Display ( DisplayFun, filterArgs, filterDisplay
	       , filterParams, dispParams, display, dxtArgs
	       , fromBitsRGB24, dxtParams, dxt_display
	       )
import DisplayConvert ( Displayable(..))
import qualified Cse as CSE
import PP hiding ( empty )
import Maybe
import List
import Char
import Monad ( when )
import IOExts (trace)
import CompileUtils
---

newtype UI a 
  = UI { deUI :: UIState -> (a, [UIElem], UIState) }

instance Monad UI where
 return x = UI (\ st -> (x,[],st))
 (UI m) >>= f = 
   UI (\ st1 -> 
         case m st1 of
	  (x, t1, st2) -> 
	     case (deUI $ f x) st2 of
	       (v, t2, st3) -> (v, t1++t2, st3))

instance Functor UI where
  fmap f ui = ui >>= return . f

data UIState
 = UIState 
    { uiNameSupply :: NameSupply      -- the supply is needed to come up with unique
    				      -- names for the global 'UI variables.'

    , uiVarDecls   :: [(Name,Type,[Name])] 
    				      -- compiler generated decls.
    , uiDecls      :: [CCode]         -- other declarations that needs to be emitted.
    , uiInits      :: [(Name,String)] -- initial values for bindings in uiVarDecls.

    , uiBackend    :: Backend       -- what host are we generating code for - is it for an
    				    -- image filter that is to be wrapped up as a PhotoShop plugin?
				    -- (this piece of the state is only passed downwards,
				    -- so it could be factored out into an environment.)

    , uiFromHost   :: Bool          -- True => anonymous image import will be provided by 
                                    --         the host (e.g., a PhotoShop host)
				    -- The reason why this is needed is that the format
				    -- of the image supplied by the host might be different
				    -- to that which you get via the UI (and UILib) -- e.g.,
				    -- a PhotoShop host supplies a stream of RGB triples; UILib
				    -- provides BGRA.
    }

data Backend 
  = Standalone 
  | Component
  | Control
  | PhotoShopFilter 
  | DXTransform
    deriving ( Eq, Show )


{-
  UIElem describes the tree of UI elements that we
  want to instantiate at run-time.
-}
data UIElem 
 = HBox String          -- grouping label
        [UIElem]
 | VBox String          -- grouping label
        [UIElem]
 | Resizeable Bool   -- True => resizeable horizontally.
 	      Bool   -- vertically
	      UIElem
 | Slider 
 	String 		-- label
	(Double,Double)	-- bounds
	Double		-- defVal
	Name		-- variable which UI will communicate changes through.
 | ISlider
 	String		-- label
	(Int,Int)	-- bounds
	Int		-- default value
	Name		-- variable which UI will communicate changes through
 | IImage
 	String		-- label
	(Maybe String)	-- default image
	(Maybe Name)	-- variable which UI will communicate image width through.
	(Maybe Name)	-- ditto, for height.
	(Maybe Name)    -- ditto, for inter-row padding.
	(Maybe Name)    -- ditto, for ptr to image data (i.e., ref to an RGBQUAD array.)
	                -- Nothing => not interested.
 | TextBox
        String          -- label
	String          -- initial text
	Name            -- variable which UI will communucate image width through.
	Name            -- ditto for height.
	Name            -- ditto for ptr to image data.
 | RadioBox
        String		-- label
	[String]	-- values
	Int		-- default value
	Name		-- variable which UI will communicate change in selection through.
 | Display
	Int Int		 -- initial size
	Name		 -- name of Displayer function
	Bool		 -- is it time-varying?
	   -- the code for the display proc.
	( [(Name,Type,[Name])]  -- externally bound vars 
	, [(Name,Type)]  -- parameters
	, DExp.Exp       -- body
	)
 | Tick
 	String		-- label
	Int		-- default value
	Name		-- UI communicates changes in status through this variable.
 | Button
 	String		-- label
	Int		-- value to report.
 | MouseInput
	Name            -- UI communicates current X position through this variable.
	Name            -- ditto for Y.
-- Disabled; only for debugging.
-- deriving (Show)


type Name     = String
type Label    = String
type FileName = String

returnUI :: UIElem -> a -> UI a
returnUI elt v =  UI $ \ st -> (v, [elt], st)

newName :: UI Name
newName = UI $ \ st -> let (x:xs) = uiNameSupply st in (x,[],st{uiNameSupply=xs})

forBackend :: UI Backend
forBackend = UI $ \ st -> (uiBackend st, [], st)

setImageFromHost :: Bool -> UI ()
setImageFromHost x = UI $ \ st -> ((), [], st{uiFromHost=x})

getImageFromHost :: UI Bool
getImageFromHost = UI $ \ st -> (uiFromHost st, [], st)

getVarDecls :: UI [(Name,Type,[Name])]
getVarDecls = UI $ \ st -> (uiVarDecls st, [], st)

addDecl :: CCode -> UI ()
addDecl d = UI $ \ st -> ((), [], st{uiDecls=d:uiDecls st})

addVarDecl :: String -> Type  -> [Name] -> UI ()
addVarDecl nm ty deps = UI $ \ st -> ((), [], st{uiVarDecls=(nm,ty,deps):uiVarDecls st})

addInit :: (Name, String) -> UI ()
addInit d = UI $ \ st -> ((), [], st{uiInits=d:uiInits st})

newUIVar :: Type -> String -> Maybe [Name] -> UI Name
newUIVar ty init delayedUpdaters = 
  UI $ \st -> 
    let
     (x:xs) = uiNameSupply st
     delayedUpdaters' =
       case delayedUpdaters of
         Nothing -> []
	 Just ls -> ls++[x]
    in
    (x, [], st{uiNameSupply=xs, uiVarDecls=(x,ty,delayedUpdaters'):uiVarDecls st, uiInits=(x,init):uiInits st})

newPropVar :: (Show a)
	   => Label
	   -> Type
	   -> a
	   -> UI Name
newPropVar lab ty defVal = do
  forB <- forBackend
  case forB of
    DXTransform -> do
      let
       propName = toCId lab

       prop =
     	 case lab of
	   "Quality" -> "m_f"++propName  -- unfortunate special case.
	   _ -> "m_" ++ propName
      addVarDecl prop ty []
      addInit (prop, show defVal)
      return prop
    Control -> do
      let
       propName = toCId lab
       prop = "m_" ++ propName
	-- in case of Controls, both the UI *and* the outside can modify
	-- the value of properties, so to catch the UI case, we need to make
	-- the property dependent on itself ==> when the property value is updated,
	-- it gets propagated the same way as other UI controls.
      addVarDecl prop ty [prop]
      addInit (prop, show defVal)
      return prop

    _ -> newUIVar ty (show defVal) (Just [])
  

mapUIElem :: (UIElem -> UIElem) -> UIElem -> UIElem
mapUIElem f elt = 
  case elt of
   HBox lab ts -> f (HBox lab (map (mapUIElem f) ts))
   VBox lab ts -> f (VBox lab (map (mapUIElem f) ts))
   Resizeable h v u -> f (Resizeable h v (f u))
   x           -> f x

mapAccumL_UIElem :: (UIElem -> a -> (a, Maybe UIElem))
	         -> a
	         -> [UIElem]
	         -> (a, [UIElem])
mapAccumL_UIElem f st []     = (st, [])
mapAccumL_UIElem f st (x:xs) =
  let
   (st2, xs') = mapAccumL_UIElem f st1 xs
   (st1, fx)  = 
     case x of 
       HBox lab ts -> 
         let
           (st1, ts') = mapAccumL_UIElem f st ts
	 in
	 (st1, \ xs' -> (HBox lab ts' : xs'))
       VBox lab ts -> 
         let
           (st1, ts') = mapAccumL_UIElem f st ts
	 in
	 (st1, \ xs' -> (VBox lab ts' : xs'))
       Resizeable h v u -> 
         case f u st of
           (st1, Nothing) -> (st1, id)
	   (st1, Just u') -> (st1, \ xs' -> (Resizeable h v u' : xs'))
       _ -> 
         case f x st of
           (st1, Nothing) -> (st1, id)
	   (st1, Just x') -> (st1, (x':))

  in
  (st2, fx xs')

---- Building blocks

-- Real-valued slider UI, given label, bounds, and default value.
-- Maybe also specify horizontal vs vertical.
slider :: Label -> (Double, Double) -> Double -> UI FloatE
slider lab bounds@(min,_) defV = do
  v   <- newPropVar lab Double defV
  returnUI (Slider lab bounds defV v) (var v)

-- A slider; fits DXTransform vocab a little better than 'slider.'
property :: String -> (Double, Double) -> Double -> UI FloatE
property = slider

-- Almost what we want for an int slider.  The only problem is that the
-- type-in box shows float values.
islider :: Label -> (Int, Int) -> Int -> UI IntE
islider lab bounds defV = do
  v    <- newPropVar lab Int defV
  returnUI (ISlider lab bounds defV v) (var v)


-- Image-valued file-browser UI, given label and default image file.  This
-- one also gives the width and height as first two return values.
browseImageSize :: Label -> FileName -> UI (FloatE,FloatE,ImageC)
browseImageSize lab fName = do
   width  <- newUIVar Int "0" Nothing
   height <- newUIVar Int "0" Nothing
   bits   <- newUIVar Int "0" (Just [width,height])
   let vWidth = var width; vHeight = var height
   returnUI (IImage lab (Just fName) (Just width) (Just height) Nothing (Just bits))
   	    (i2f vWidth, i2f vHeight,
             (reconstruct (Array2 vWidth vHeight
			   (sampleBGR32 (var bits) vWidth))))

-- Image-valued file-browser UI, given label and default image file.
browseImage :: Label -> FileName -> UI ImageC
browseImage lab fName = do (w,h,im) <- browseImageSize lab fName
                           return im

{- 
  Unbound image imports - depending on the host you're targetting,
  the image import could either be provided by the host directly (e.g.,
  the image you filter in a Photoshop plugin) or as an UI element.
 
  In the case where an image import can't be bound by the host *and* it
  is not possible to bind it via an UI element, it is a static error.
  However, this is not a problem for any of Pan's current targets.

-}

-- Discrete import
imageSourceD :: Label -> UI (Array2 ColorE)
imageSourceD lab = do
   width  <- newUIVar Int "0" Nothing
   height <- newUIVar Int "0" Nothing
   bits   <- newUIVar Int "0" (Just [width,height])
   forB   <- forBackend
   let vWidth  = var width
       vHeight = var height
   (sampler, mbPad) <-
       -- if the image source is bound by PhotoShop, it is
       -- a pointer to a sequence of RGB triples. Sample the
       -- bits pointer accordingly.
      case forB of
	   PhotoShopFilter -> do
	        flg <- getImageFromHost
		if flg then do
		  setImageFromHost False
	          pad <- newUIVar Int "0" Nothing
		  let vPad = (var pad :: IntE)
	   	  return (fromBitsRGB24 (var bits) vWidth vHeight vPad, Just pad)
		 else do
		  return (sampleBGR32 (var bits) vWidth, Nothing)
	   DXTransform     -> return (sampleBGRA32 (var bits) vWidth, Nothing)
	   _               -> return (sampleBGR32 (var bits) vWidth, Nothing)

   returnUI (IImage lab Nothing (Just width) (Just height) mbPad (Just bits))
   	    (Array2 vWidth vHeight sampler)

imageSourceSize :: Label -> UI (FloatE, FloatE, ImageC)
imageSourceSize lab = do
   bmp@(Array2 w h _) <- imageSourceD lab
   return (i2f w, i2f h, reconstruct bmp)

imageSource :: Label -> UI (ImageC)
imageSource lab = do
   (_,_,im) <- imageSourceSize lab
   return im

-- Browse image and introduce alpha.
browseImageWithAlphaSize :: Label -> FileName -> UI (FloatE,FloatE,ImageC)
browseImageWithAlphaSize lab dflt = 
  do (w,h,im) <- browseImageSize lab dflt
     addTransp <- radioBox "Transparency style"
                  [ ("color on white",
                     chromaKey 0.01 white)
                  , ("anti-aliased grey scale",
                     transpLightI)
		  ]
		  0
     return $ (w,h, addTransp im)     

browseImageWithAlpha :: Label -> FileName -> UI ImageC
browseImageWithAlpha lab fName =
  do (w,h,im) <- browseImageWithAlphaSize lab fName
     return im


-- Discrete version
textBoxD :: String -> UI (Array2 ColorE)
textBoxD lab = do
   width  <- newUIVar Int "0" Nothing
   height <- newUIVar Int "0" Nothing
   bits   <- newUIVar Int "0" (Just [width,height])
   let vWidth  = var width
       vHeight = var height
       sampler = sampleBGR32 (var bits) vWidth
   returnUI (TextBox "Text" lab width height bits)
   	    (Array2 vWidth vHeight sampler)

textBoxSize :: String -> UI (FloatE, FloatE, ImageC)
textBoxSize lab = 
  do arr@(Array2 vWidth vHeight _) <- textBoxD lab
     return (i2f vWidth, i2f vHeight, reconstruct arr)

textBox :: String -> UI ImageC
textBox lab = do
  (w,h,im) <- textBoxSize lab
  return im

radioBoxInt :: Label -> [Label] -> Int -> UI IntE
radioBoxInt lab choices defV = do
  let

     defV' 
       | defV >= length choices = max 0 ((length choices) - 1)
       | defV < 0 = 0
       | otherwise = defV
  val  <- newPropVar lab Int defV'
  returnUI (RadioBox lab choices defV' val)
  	   (var val)

-- Yield the list element corresponding to the chosen label
radioBox :: Syntactic a => String -> [(Label,a)] -> Int -> UI a
radioBox label choices defV =
  fmap (choiceVals !!*) $ radioBoxInt label choiceNames defV
 where
   (choiceNames,choiceVals) = unzip choices

-- check box, has value of '1' when checked, '0' otherwise.
tickInt :: Label -> Int -> UI IntE
tickInt lab defV = do
  let defV'
  	| defV > 1  = 1
	| defV < 0  = 0
	| otherwise = defV

  val  <- newPropVar lab Int defV'
  returnUI (Tick lab defV' val)
  	   (var val)

-- Boolean UI
tick :: Label -> Bool -> UI BoolE
tick label def = fmap (/=* 0) $ tickInt label (bool2Int def)
 where
  bool2Int True = 1
  bool2Int _    = 0

mouseInput :: UI PointE
mouseInput = do
   x <- newUIVar Double "0" (Just [])
   y <- newUIVar Double "0" (Just [])
   b <- forBackend
   if b == DXTransform || b == PhotoShopFilter then
     trace "Warning: Mouse input is unsupported for Photoshop plugins and DXTransforms; ignoring\n" $
     return (var x, var y)
    else
     returnUI (MouseInput x y)
     	      (var x, var y)

-- Apply a function-valued UI if enabled.
maybeUI :: Syntactic a => Label -> UI (a -> a) -> UI (a -> a)
maybeUI label fui = do c <- tick label False
                       f <- fui
                       return $ ifE c f id

tickFun :: Syntactic a => Label -> (a -> a) -> UI (a -> a)
tickFun label f = maybeUI label (return f)


-- buttons are only supported in order to provide Ok/Cancel functionality.
button :: Label -> Int -> UI ()
button lab val = returnUI (Button lab val) ()

-- Display sub-window, given (initial?) size and image.  No info coming
-- out for now, though we could produce something like frame rate,
-- pixels/sec, or mouse position.
mkDisplay :: (Displayable a) => (Int,Int) -> a -> UI ()
mkDisplay (w,h) img = do
  v <- newName 
  let dispProcName = "display_"++v
  v <- newName 
  ls <- getVarDecls
  let
     -- we determine whether the expression is time-varying by seeing if 
     -- 't' occurs free in the final, simplified expression.
    varying    = timeVarying exp
    displayFun = toDisplayer True img

    exp = CSE.cse (Exp.unE (display displayFun))

  returnUI (Display w h dispProcName varying (ls,dispParams,exp))
	   ()

pluginDisplay :: (Displayable a) => (Int,Int) -> a -> UI ()
pluginDisplay (w,h) f = do
  v <- newName
  let dispProcName = "display_"++v
  exp' <-
    if varying then do
       E sl <- slider "progress" (0,10) 0
       return (substVar timeVar sl exp)
     else
       return exp
  ls   <- getVarDecls
  returnUI (Display w h dispProcName varying (ls,filterParams,exp')) ()
 where
   timeVar = "t" -- Note: brittle, but this *has* to match the name of the
   		 -- 't' lambda variable used in Display.filterDisplay

   -- Note: we do have the possibility of generating a UI element
   -- that represents time here.
   varying = timeVarying exp
   f'      = toDisplayer True f
   exp     = CSE.cse (Exp.unE (filterDisplay f' filterArgs))

dxtDisplay :: (Displayable a) => (Int,Int) -> a -> UI ()
dxtDisplay (w,h) f = do
   -- Note: only *one* display supported on the DXTransform
   -- side.
  let dispProcName = "CPanImageImpl::Display"
  exp' <-
    if varying then do
       E sl <- slider "Time" (0,10) 0
       return (substVar "t" sl exp)
     else
       return exp
  ls   <- getVarDecls
  returnUI (Display w h dispProcName False (ls,dxtParams,exp')) ()
 where
   varying = timeVarying exp

   f'      = toDisplayer False f
   exp     = CSE.cse (Exp.unE (dxt_display f' dxtArgs))

-- For display, if we have mouse position coming out, it would really be
-- useful to feed it back into the ImageC.  Maybe display sub-window size
-- also.  Here's a possible tool, in analogy to fixIO.
-- fixUI :: (a -> UI a) -> UI a

-- Layout: horizontal and vertical.
hbox, vbox :: UI a -> UI a
hbox (UI act) = UI $ \ st ->
  case act st of
   (v, ts, st2) -> 
	case removeEmptyBoxes ts of
	  [] -> (v, [], st2)
	  [x] -> (v, [x], st2)
	  xs  -> (v, [HBox "" xs], st2)
  
vbox (UI act) = UI $ \ st ->
  case act st of
   (v, ts, st2) -> 
   	case removeEmptyBoxes ts of
	  [] -> (v,[],st2)
	  [x] -> (v, [x], st2)
	  xs -> (v, [VBox "" xs], st2)

fixedH :: UI a -> UI a
fixedH (UI a) = UI $ \ st -> 
	case a st of
	 (v,ts,st2) -> (v, map (Resizeable False True) ts, st2)

fixedV :: UI a -> UI a
fixedV (UI a) = UI $ \ st -> 
	case a st of
	 (v,ts,st2) -> (v, map (Resizeable True False) ts, st2)

{-
  In the case where no explicit layout has been used (at the toplevel),
  you might find the following transformation useful:
    * lay out the components vertically
    * any child containing a 'Display' is lifted to the top.

  The motivation for this transformation is based on the view that
  it is preferable to have the interactive controls underneath the
  display(s). 
-}
topBox :: UI a -> UI a
topBox (UI act) = UI $ \ st ->
  case act st of
   (v, ts, st2) -> (v, [topBoxLayout ts], st2)

topBoxLayout :: [UIElem] -> UIElem
topBoxLayout ts =
        let
         (ds, no_ds) = partition isDisp ts
        in
	VBox "" (ds ++ no_ds)

okCancel :: UI a -> UI a
okCancel ui = 
   hbox $ do
     x <- topBox ui
     fixedH $ vbox $ do
	  button "Ok" 1
	  button "Cancel" 0
     return x

isDisp :: UIElem -> Bool
isDisp Display{} = True
isDisp _	 = False

removeEmptyBoxes :: [UIElem] -> [UIElem]
removeEmptyBoxes [] = []
removeEmptyBoxes ((HBox _ []):xs) = removeEmptyBoxes xs
removeEmptyBoxes ((VBox _ []):xs) = removeEmptyBoxes xs
removeEmptyBoxes (x:xs) = x : removeEmptyBoxes xs

-- ToDo: remove these two out into some PP support module.
ppDecl nm ty init = text (showCType ty) <+> text nm <> equals <> text init <> semi

toCId :: String -> String
toCId ls = map toIdChar ls
 where
    toIdChar c
      | not (isAlphaNum c) = '_'
      | otherwise = c


{-
  When compiling up an effect, we need to determine whether
  the final expression/program is time-dependent or not. We determine
  this by checking to see whether (Lambda "t" Double e) occurs
  & if 't' is actually used in e.
  
  Note: the use of 't' has to match up with the name used for the time
  lambda-var used by Display.display and Display.filterDisplay.
-}
timeVarying :: DExp.Exp -> Bool
timeVarying (DExp.Let _ _ e) = timeVarying e
timeVarying (DExp.Lambda "t" Double e) = "t" `DExp.occFreeIn` e
timeVarying (DExp.Lambda _ _ e) = timeVarying e
timeVarying e = False


